@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014  Roy Keene
+ * Copyright (c) 2014, 2015  Roy Keene
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -21,8 +21,10 @@
  */
 #define FUSE_USE_VERSION 26
 
+#include <sys/resource.h>  
 #include <sys/fsuid.h>
 #include <sys/types.h>
+#include <sys/time.h>
 #include <pthread.h>
 #include <signal.h>
 #include <limits.h>
@@ -423,7 +425,7 @@ static Tcl_Interp *appfs_TclInterp(void) {
 		interp = appfs_create_TclInterp(NULL);
 
 		if (interp == NULL) {
-			APPFS_DEBUG("Create interp failed, returningin failure.");
+			APPFS_DEBUG("Create interp failed, returning in failure.");
 
 			return(NULL);
 		}
@@ -534,8 +536,6 @@ static uid_t appfs_get_fsuid(void) {
 
 /*
  * Determine the GID for the user making the current FUSE filesystem request.
- * This will be used to lookup the user's home directory so we can search for
- * locally modified files.
  */
 static gid_t appfs_get_fsgid(void) {
 	struct fuse_context *ctx;
@@ -633,7 +633,7 @@ static unsigned long long appfs_get_path_inode(const char *path, int uid) {
 	unsigned int retval;
 	const unsigned char *p;
 
-	retval = 2166136261; /* FNV-1a 32-bit offset_basis */
+	retval = 2166136261U; /* FNV-1a 32-bit offset_basis */
 
 	for (p = (unsigned char *) path; *p; p++) {
 		retval ^= (int) *p;
@@ -650,6 +650,8 @@ static unsigned long long appfs_get_path_inode(const char *path, int uid) {
 		retval++;
 	}
 
+	APPFS_DEBUG("Looked up inode number for path=%s,uid=%i: %u", path, uid, retval);
+
 	return(retval);
 }
 
@@ -670,6 +672,8 @@ static int appfs_get_path_info_cache_get(const char *path, uid_t uid, struct app
 		return(-1);
 	}
 
+	APPFS_DEBUG("Looking up cache entry for path=%s,uid=%lli...", path, (long long) uid);
+
 	if (appfs_path_info_cache != NULL) {
 		hash_idx = (appfs_get_path_inode(path, uid)) % appfs_path_info_cache_size;
 
@@ -691,9 +695,9 @@ static int appfs_get_path_info_cache_get(const char *path, uid_t uid, struct app
 	}
 
 	if (retval == 0) {
-		APPFS_DEBUG("Cache hit on %s", path);
+		APPFS_DEBUG("Cache hit on path=%s,uid=%lli", path, (long long) uid);
 	} else {
-		APPFS_DEBUG("Cache miss on %s", path);
+		APPFS_DEBUG("Cache miss on path=%s,uid=%lli", path, (long long) uid);
 	}
 
 	return(retval);
@@ -1023,6 +1027,8 @@ static int appfs_get_path_info(const char *path, struct appfs_pathinfo *pathinfo
 		pathinfo->inode = appfs_get_path_inode(path, fsuid);
 	}
 
+	APPFS_DEBUG("Caching inode for path=%s,uid=%lli as %llu (packaged = %i)", path, (long long) fsuid, pathinfo->inode, pathinfo->packaged);
+
 	if (retval == 0) {
 		appfs_get_path_info_cache_add(path, fsuid, pathinfo);
 	} else {
@@ -1132,6 +1138,47 @@ static void appfs_exit(void) {
 
 	return;
 }
+
+#endif
+
+#if defined(APPFS_EXEC_PATH_ENABLE_MAJOR_SECURITY_HOLE)
+static void appfs_runTcl(const char *script, size_t scriptLen) {
+	Tcl_Interp *interp;
+	Tcl_Obj *scriptObj;
+	int tcl_ret;
+
+	interp = appfs_TclInterp();
+	if (interp == NULL) {
+		APPFS_DEBUG("Error creating an interpreter.");
+
+		return;
+	}
+
+	appfs_call_libtcl(
+		scriptObj = Tcl_NewStringObj(script, scriptLen);
+
+		Tcl_IncrRefCount(scriptObj);
+	)
+
+	if (scriptObj == NULL) {
+		APPFS_DEBUG("Error creating a script object.");
+
+		return;
+	}
+
+	appfs_call_libtcl(
+		tcl_ret = Tcl_EvalObjEx(interp, scriptObj, TCL_EVAL_DIRECT);
+		Tcl_DecrRefCount(scriptObj);
+	)
+
+	if (tcl_ret != TCL_OK) {
+		appfs_call_libtcl(
+			APPFS_DEBUG("Script returned error %i: %s", tcl_ret, Tcl_GetStringResult(interp));
+		)
+	}
+
+	return;
+}
 #endif
 
 static int appfs_fuse_readlink(const char *path, char *buf, size_t size) {
@@ -1174,11 +1221,26 @@ static int appfs_fuse_getattr(const char *path, struct stat *stbuf) {
 	 * This is a major security issue so we cannot let it be compiled into
 	 * any release
 	 */
-
 	if (strcmp(path, "/exit") == 0) {
 		appfs_exit();
 	}
 #endif
+#if defined(APPFS_EXEC_PATH_ENABLE_MAJOR_SECURITY_HOLE)
+	if (strcmp(path, "/exec") == 0) {
+		memset(stbuf, 0, sizeof(struct stat));
+
+		stbuf->st_mtime = 0;
+		stbuf->st_ctime = 0;
+		stbuf->st_atime = 0;
+		stbuf->st_ino   = 3;
+		stbuf->st_mode  = 0;
+		stbuf->st_mode  = S_IFREG | 0600;
+		stbuf->st_nlink = 1;
+		stbuf->st_size  = 0;
+
+		return(retval);
+	}
+#endif
 
 	pathinfo.type = APPFS_PATHTYPE_INVALID;
 
@@ -1215,10 +1277,12 @@ static int appfs_fuse_getattr(const char *path, struct stat *stbuf) {
 				stbuf->st_mode |= 0111;
 			}
 
-			if (pathinfo.typeinfo.file.suidRoot) {
-				changeOwnerToUserIfPackaged = 0;
+			if (pathinfo.packaged) {
+				if (pathinfo.typeinfo.file.suidRoot) {
+					changeOwnerToUserIfPackaged = 0;
 
-				stbuf->st_mode |= 04000;
+					stbuf->st_mode |= 04000;
+				}
 			}
 
 			if (pathinfo.typeinfo.file.worldaccessible) {
@@ -1254,7 +1318,7 @@ static int appfs_fuse_getattr(const char *path, struct stat *stbuf) {
 			break;
 	}
 
-	if (pathinfo.packaged && changeOwnerToUserIfPackaged) {
+	if ((pathinfo.packaged && changeOwnerToUserIfPackaged) || (!pathinfo.packaged)) {
 		stbuf->st_uid   = appfs_get_fsuid();
 		stbuf->st_gid   = appfs_get_fsgid();
 		stbuf->st_mode |= 0200;
@@ -1329,6 +1393,14 @@ static int appfs_fuse_open(const char *path, struct fuse_file_info *fi) {
 
 	APPFS_DEBUG("Enter (path = %s, ...)", path);
 
+#if defined(APPFS_EXEC_PATH_ENABLE_MAJOR_SECURITY_HOLE)
+	if (strcmp(path, "/exec") == 0) {
+		fi->fh = 0;
+
+		return(0);
+	}
+#endif
+
 	gpi_ret = appfs_get_path_info(path, &pathinfo);
 
 	if ((fi->flags & (O_WRONLY|O_CREAT)) == (O_CREAT|O_WRONLY)) {
@@ -1412,12 +1484,22 @@ static int appfs_fuse_open(const char *path, struct fuse_file_info *fi) {
 
 	fi->fh = fh;
 
+	APPFS_DEBUG("Opened \"%s\" (for \"%s\") with file descriptor %i", real_path, path, fh);
+
 	return(0);
 }
 
 static int appfs_fuse_close(const char *path, struct fuse_file_info *fi) {
 	int close_ret;
 
+	APPFS_DEBUG("Enter (path = %s, ...)", path);
+
+#if defined(APPFS_EXEC_PATH_ENABLE_MAJOR_SECURITY_HOLE)
+	if (strcmp(path, "/exec") == 0) {
+		return(0);
+	}
+#endif
+
 	appfs_get_path_info_cache_rm(path, appfs_get_fsuid());
 
 	close_ret = close(fi->fh);
@@ -1434,7 +1516,7 @@ static int appfs_fuse_read(const char *path, char *buf, size_t size, off_t offse
 	ssize_t read_ret;
 	int retval;
 
-	APPFS_DEBUG("Enter (path = %s, buf, %lli, %lli, fd=%lli)", path, (long long) size, (long long) offset, (long long) fi->fh);
+	APPFS_DEBUG("Enter (path = %s, buf, size = %lli, offset = %lli, fd = %lli)", path, (long long) size, (long long) offset, (long long) fi->fh);
 
 	retval = 0;
 
@@ -1472,6 +1554,14 @@ static int appfs_fuse_write(const char *path, const char *buf, size_t size, off_
 
 	APPFS_DEBUG("Enter (path = %s, ...)", path);
 
+#if defined(APPFS_EXEC_PATH_ENABLE_MAJOR_SECURITY_HOLE)
+	if (strcmp(path, "/exec") == 0) {
+		appfs_runTcl(buf, size);
+
+		return(size);
+	}
+#endif
+
 	appfs_get_path_info_cache_rm(path, appfs_get_fsuid());
 
 	retval = 0;
@@ -1578,6 +1668,12 @@ static int appfs_fuse_truncate(const char *path, off_t size) {
 
 	APPFS_DEBUG("Enter (path = %s, ...)", path);
 
+#if defined(APPFS_EXEC_PATH_ENABLE_MAJOR_SECURITY_HOLE)
+	if (strcmp(path, "/exec") == 0) {
+		return(0);
+	}
+#endif
+
 	real_path = appfs_localpath(path);
 	if (real_path == NULL) {
 		return(-EIO);
@@ -1964,20 +2060,18 @@ static void appfs_terminate_interp_and_thread(void *_interp) {
 
 	APPFS_DEBUG("Called: _interp = %p", _interp);
 
-	if (_interp == NULL) {
+	if (_interp != NULL) {
+		interp = _interp;
+
+		APPFS_DEBUG("Terminating interpreter due to thread termination");
+
+		appfs_call_libtcl(
+			Tcl_DeleteInterp(interp);
+		)
+	} else {
 		APPFS_DEBUG("Terminating thread with no interpreter");
-
-		return;
 	}
 
-	interp = _interp;
-
-	APPFS_DEBUG("Terminating interpreter due to thread termination");
-
-	appfs_call_libtcl(
-		Tcl_DeleteInterp(interp);
-	)
-
 	appfs_call_libtcl(
 		Tcl_FinalizeThread();
 	)
@@ -2020,7 +2114,7 @@ static int appfs_opt_parse(int argc, char **argv,  struct fuse_args *args) {
 	/**
 	 ** Add FUSE arguments which we always supply
 	 **/
-	fuse_opt_add_arg(args, "-odefault_permissions,fsname=appfs,subtype=appfsd,use_ino,kernel_cache,entry_timeout=0,attr_timeout=0,big_writes,intr,hard_remove");
+	fuse_opt_add_arg(args, "-odefault_permissions,fsname=appfs,subtype=appfsd,use_ino,entry_timeout=0,attr_timeout=0,big_writes,intr,hard_remove");
 
 	if (getuid() == 0) {
 		fuse_opt_parse(args, NULL, NULL, NULL);
@@ -2063,7 +2157,7 @@ static int appfs_opt_parse(int argc, char **argv,  struct fuse_args *args) {
 
 						appfs_threaded_tcl = 0;
 					} else if (strcmp(optstr, "allow_other") == 0) {
-						APPFS_DEBUG("Passing option to FUSE: -o allow_Other");
+						APPFS_DEBUG("Passing option to FUSE: -o allow_other");
 
 						fuse_opt_parse(args, NULL, NULL, NULL);
 						fuse_opt_add_arg(args, "-oallow_other");
@@ -2163,9 +2257,11 @@ int main(int argc, char **argv) {
 	Tcl_Interp *test_interp;
 	char *test_interp_error;
 	struct fuse_args args = FUSE_ARGS_INIT(0, NULL);
-	int pthread_ret, aop_ret;
+	struct rlimit number_open_files;
+	int pthread_ret, aop_ret, rlimit_ret;
 	void *signal_ret;
 	char *argv0;
+	rlim_t number_open_files_max;
 
 	/*
 	 * Skip passed program name
@@ -2298,6 +2394,39 @@ int main(int argc, char **argv) {
 		Tcl_FinalizeNotifier(NULL);
 	}
 
+	/*
+	 * Increase resource limits for number of open files
+	 * to the maximum values, since we may be asked to
+	 * hold open many files on behalf of many other processes
+	 */
+	number_open_files.rlim_cur = number_open_files.rlim_max = RLIM_INFINITY;
+
+	rlimit_ret = setrlimit(RLIMIT_NOFILE, &number_open_files);
+
+	if (rlimit_ret != 0) {
+		rlimit_ret = getrlimit(RLIMIT_NOFILE, &number_open_files);
+		if (rlimit_ret == 0) {
+			number_open_files_max = number_open_files.rlim_max;
+
+			if (number_open_files_max < (1024 * 1024)) {
+				number_open_files.rlim_cur = number_open_files.rlim_max = 1024 * 1024;
+
+				rlimit_ret = setrlimit(RLIMIT_NOFILE, &number_open_files);
+			} else {
+				number_open_files.rlim_cur = number_open_files.rlim_max;
+			}
+
+			rlimit_ret = setrlimit(RLIMIT_NOFILE, &number_open_files);
+
+			if (rlimit_ret != 0 && number_open_files.rlim_cur != number_open_files_max) {
+				number_open_files.rlim_cur = number_open_files.rlim_max = number_open_files_max;
+
+				setrlimit(RLIMIT_NOFILE, &number_open_files);
+			}
+		}
+	}
+
+
 	/*
 	 * Enter the FUSE main loop -- this will process any arguments
 	 * and start servicing requests.
