@@ -31,6 +31,8 @@ package require pki
 
 # Functions specifically meant for users to replace as a part of configuration
 namespace eval ::appfs::user {
+	variable download_method "tcl"
+
 	# User-replacable function to convert a hostname/hash/method to an URL
 	proc construct_url {hostname hash method} {
 		return "http://$hostname/appfs/$method/$hash"
@@ -53,12 +55,59 @@ namespace eval ::appfs::user {
 
 		return $perms
 	}
+
+	# User-replacable function to fetch a remote file
+	proc download_file {url {outputChannel ""}} {
+		switch -- $::appfs::user::download_method {
+			"curl" {
+				if {$outputChannel eq ""} {
+					return [exec curl -sS -L -- $url]
+				} else {
+					exec curl -sS -L -- $url >@ $outputChannel
+
+					return ""
+				}
+			}
+			"tcl" {
+				catch {
+					if {$outputChannel eq ""} {
+						set token [http::geturl $url]
+						set retval [http::data $token]
+					} else {
+						set token [http::geturl $url -binary true -channel $outputChannel]
+						set retval ""
+					}
+				} err
+
+				if {![info exists token]} {
+					return -code error "Unable to download \"$url\": $err"
+				}
+
+				set tokenCode [http::ncode $token]
+
+				http::cleanup $token
+
+				if {$tokenCode != "200"} {
+					return -code error "Unable to download \"$url\": Site did not return a 200 (returned $tokenCode)"
+				}
+
+				if {![info exists retval]} {
+					return -code error "Unable to download \"$url\": Site did not return proper data: $err"
+				}
+
+				return $retval
+			}
+
+		}
+
+		return -code error "Unable to download"
+	}
 }
 
 namespace eval ::appfs {
 	variable cachedir "/tmp/appfs-cache"
 	variable ttl 3600
-	variable nttl 60
+	variable nttl 3600
 	variable trusted_cas [list]
 	variable platform [::platform::generic]
 
@@ -71,7 +120,11 @@ namespace eval ::appfs {
 		return $retval
 	}
 
-	proc _cachefile {url key {keyIsHash 1}} {
+	proc _cachefile {url key method {keyIsHash 1}} {
+		if {$keyIsHash && $method != "sha1"} {
+			return -code error "Only SHA1 hashing method is supported"
+		}
+
 		set filekey $key
 		if {$keyIsHash} {
 			set filekey [_hash_sep $filekey]
@@ -91,14 +144,7 @@ namespace eval ::appfs {
 		fconfigure $fd -translation binary
 
 		catch {
-			set token [::http::geturl $url -channel $fd -binary true]
-		}
-
-		if {[info exists token]} {
-			set ncode [::http::ncode $token]
-			::http::reset $token
-		} else {
-			set ncode "900"
+			::appfs::user::download_file $url $fd
 		}
 
 		close $fd
@@ -109,7 +155,7 @@ namespace eval ::appfs {
 			set hash $key
 		}
 
-		if {$ncode == "200" && $hash == $key} {
+		if {$hash == $key} {
 			file rename -force -- $tmpfile $file
 		} else {
 			file delete -force -- $tmpfile
@@ -278,7 +324,7 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 
 	proc download {hostname hash {method sha1}} {
 		set url [::appfs::user::construct_url $hostname $hash $method]
-		set file [_cachefile $url $hash]
+		set file [_cachefile $url $hash $method]
 
 		if {![file exists $file]} {
 			return -code error "Unable to fetch (file does not exist: $file)"
@@ -288,6 +334,10 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 	}
 
 	proc getindex {hostname} {
+		if {[string match "*\[/~\]*" $hostname]} {
+			return -code error "Invalid hostname"
+		}
+
 		set now [clock seconds]
 
 		set lastUpdates [db eval {SELECT lastUpdate, ttl FROM sites WHERE hostname = $hostname LIMIT 1;}]
@@ -303,25 +353,17 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 			return COMPLETE
 		}
 
-		if {[string match "*\[/~\]*" $hostname]} {
-			return -code error "Invalid hostname"
-		}
-
 		set url "http://$hostname/appfs/index"
 
 		catch {
-			set token [::http::geturl $url]
-			if {[::http::ncode $token] == "200"} {
-				set indexhash_data [::http::data $token]
-			}
-			::http::reset $token
-			::http::cleanup $token
+			set indexhash_data [::appfs::user::download_file $url]
 		}
 
-		if {![info exists indexhash_data]} {
-			# Cache this result for 60 seconds
-			db eval {INSERT OR REPLACE INTO sites (hostname, lastUpdate, ttl) VALUES ($hostname, $now, $::appfs::nttl);}
+		# Note that we attempted to fetch this index and do not try
+		# again for a while
+		db eval {INSERT OR REPLACE INTO sites (hostname, lastUpdate, ttl) VALUES ($hostname, $now, $::appfs::nttl);}
 
+		if {![info exists indexhash_data]} {
 			return -code error "Unable to fetch $url"
 		}
 
@@ -341,8 +383,19 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 		}
 
 		set file [download $hostname $indexhash]
-		set fd [open $file]
-		set data [read $fd]
+		catch {
+			set fd [open $file]
+		}
+
+		if {![info exists fd]} {
+			return -code error "Unable to download or open $file"
+		}
+
+		unset -nocomplain data
+		catch {
+			set data [read $fd]
+		}
+
 		close $fd
 
 		set curr_packages [list]
@@ -424,8 +477,19 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 		}
 
 		set file [download $hostname $package_sha1]
-		set fd [open $file]
-		set pkgdata [read $fd]
+
+		catch {
+			set fd [open $file]
+		}
+
+		if {![info exists fd]} {
+			return -code error "Unable to download or open $file"
+		}
+
+		catch {
+			set pkgdata [read $fd]
+		}
+
 		close $fd
 
 		db transaction {
@@ -679,6 +743,8 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 			::appfs::getpkgmanifest $pathinfo(hostname) $pathinfo(package_sha1)
 		}
 
+		set retval(path_type) $pathinfo(_type)
+
 		switch -- $pathinfo(_type) {
 			"toplevel" {
 				set retval(type) directory
@@ -756,6 +822,7 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 
 				if {$localpath != "" && [info exists localpathinfo]} {
 					set retval(is_localfile) 1
+					unset retval(packaged)
 					catch {
 						set retval(time) $localpathinfo(mtime)
 
@@ -841,7 +908,7 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 		array set pathinfo [_parsepath $path]
 
 		if {$pathinfo(_type) != "files"} {
-			return -code error "invalid type"
+			return -code error "invalid path type: Got \"$pathinfo(_type)\", need \"files\""
 		}
 
 		set localpath [_localpath $pathinfo(package) $pathinfo(hostname) $pathinfo(file)]
@@ -905,7 +972,7 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 		array set pathinfo [_parsepath $path]
 
 		if {$pathinfo(_type) != "files"} {
-			return -code error "invalid type"
+			return -code error "invalid path type: Got \"$pathinfo(_type)\", need \"files\""
 		}
 
 		set localpath [_localpath $pathinfo(package) $pathinfo(hostname) $pathinfo(file)]
@@ -954,8 +1021,8 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 	proc unlinkpath {path} {
 		array set pathattrs [exists $path]
 
-		if {![info exists pathattrs(packaged)]} {
-			return -code error "invalid type"
+		if {$pathattrs(path_type) != "files"} {
+			return -code error "invalid path type: can only delete type \"files\" this is type \"$pathattrs(path_type)\""
 		}
 
 		set localpath $pathattrs(localpath)
